from coalib.bearlib.languages.Language import Language


@Language
class Python:
    aliases = 'py',
    versions = 2.7, 3.3, 3.4, 3.5, 3.6

    extensions = '.py',
    comment_delimiter = '#'
    multiline_comment_delimiters = {}
    string_delimiters = {'"': '"', "'": "'"}
    multiline_string_delimiters = {'"""': '"""', "'''": "'''"}
    indent_types = ':',
    encapsulators = {'(': ')', '[': ']', '{': '}'}



import logging
import os

from coalib.output.Logging import configure_logging
from coalib.parsing import Globbing
from coalib.settings.ConfigurationGathering import get_config_directory
from coalib.settings.Section import Section
from coalib.parsing.Globbing import glob_escape


def main(log_printer=None, section: Section=None):
    configure_logging()

    start_path = get_config_directory(section)

    if start_path is None:
        return 255

    # start_path may have unintended glob characters
    orig_files = Globbing.glob(os.path.join(
        glob_escape(start_path), '**', '*.orig'))

    not_deleted = 0
    for ofile in orig_files:
        logging.info('Deleting old backup file... ' + os.path.relpath(ofile))
        try:
            os.remove(ofile)
        except OSError as oserror:
            not_deleted += 1
            logging.warning("Couldn't delete {}. {}".format(
                os.path.relpath(ofile), oserror.strerror))

    if not_deleted:
        logging.warning(str(not_deleted) + ' .orig backup files could not be'
                                           ' deleted, possibly because you '
                                           'lack the permission to do so. '
                                           'coala may not be able to create'
                                           ' backup files when patches are '
                                           'applied.')
    return 0


if __name__ == '__main__':  # pragma: no cover
    main()



"""
This directory holds means to get generic information for specific languages.
"""

# Start ignoring PyUnusedCodeBear
from .Language import Language
from .Language import Languages

from .definitions.Unknown import Unknown
from .definitions.C import C
from .definitions.CPP import CPP
from .definitions.CSharp import CSharp
from .definitions.CSS import CSS
from .definitions.Fortran import Fortran
from .definitions.Golang import Golang
from .definitions.html import HTML
from .definitions.Java import Java
from .definitions.JavaScript import JavaScript
from .definitions.JSP import JSP
from .definitions.Matlab import Matlab
from .definitions.ObjectiveC import ObjectiveC
from .definitions.PHP import PHP
from .definitions.PLSQL import PLSQL
from .definitions.Python import Python
from .definitions.Ruby import Ruby
from .definitions.Scala import Scala
from .definitions.Swift import Swift
from .definitions.Vala import Vala
# Stop ignoring PyUnusedCodeBear




from functools import wraps

from .taste import Taste
from .meta import aspectclass
from coalib.settings.FunctionMetadata import FunctionMetadata


def map_setting_to_aspect(**aspectable_setting):
    """
    Map function arguments with aspect and override it if appropriate.

    This decorator can be used by ``Bear.run()`` to automatically map and
    override bear's setting value with their equivalent aspect or taste.

    The order of setting override from the lowest to highest is:
    - Setting default (in bear's run argument)
    - Aspect/taste default (if aspect is activated in Section)
    - Explicit aspect/taste default (if aspect is activated in Section)
    - Explicit setting

    :param aspectable_setting:
        A dictionary of settings as keys and their equivalent aspect or taste
        as value.
    """
    def _func_decorator(func):
        @wraps(func)
        def _new_func(self, *args, **kwargs):
            if self.section.aspects:
                aspects = self.section.aspects
                for arg, aspect_value in aspectable_setting.items():
                    # Explicit setting takes priority
                    if arg in self.section:
                        continue

                    if isinstance(aspect_value, aspectclass):
                        kwargs[arg] = aspects.get(aspect_value) is not None
                    if isinstance(aspect_value, Taste):
                        aspect_instance = aspects.get(aspect_value.aspect_name)
                        if aspect_instance:
                            kwargs[arg] = aspect_instance.tastes[
                                aspect_value.name]

            return func(self, *args, **kwargs)

        # Keep metadata
        _new_func.__metadata__ = FunctionMetadata.from_function(func)

        return _new_func

    return _func_decorator



from coalib.bearlib.aspects import Taste, Root


@Root.subaspect
class Smell:
    """
    Symptom in a piece of code that possibly indicates a deeper problem.

    `Smells` are certain structures in a code that indicate violation of
    fundamental design principles. They are usually not bugs; they are not
    technically incorrect and do not currently prevent the program from
    functioning.
    """
    class docs:
        example = """
        =begin
        Example of Ruby code with data clumps and methods with too many
        parameters.
        =end
        class Dummy
            def x(y1, y2, y3, y4, y5, y6, y7, y8, a); end
            def y(y1, y2, y3, y4, y5, y6, y7, y8); end
            def z(y1, y2, y3, y4, y5, y6, y7, y8); end
        end
        """
        example_language = 'Ruby'
        importance_reason = """
        Even though they are not necessarily bugs, code smells increase the
        risk of bugs or failure in the future and may slow down development.
        """
        fix_suggestions = """
        There are several `refactoring techniques` that can be used to deal
        with `code smells` including:

        * Composing methods
        * Moving features between objects
        * Organizing data
        * Simplifying conditional expressions
        * Simplifying method calls
        * Dealing with generalisation

        See <https://sourcemaking.com/refactoring/refactorings> for more
        information.
        """


@Smell.subaspect
class ClassSmell:
    """
    Code smells related to classes' definition.

    Class-level code smells indicate poorly defined classes (including too
    large classes or God object, data clump feature envy etc...) in your
    source code.
    """
    class docs:
        example = """
        class Warehouse
            def sale_price(item)
                item.price - item.rebate
            end
        end

        # sale_price refers to item more than self.
        """
        example_language = 'Ruby'
        importance_reason = """
        These classes should be refactored for better readability and
        maintainability of your source code.
        """
        fix_suggestions = """
        When a class is wearing too many (functional) hats (too large
        classes), you should probably think about splitting it up:

        * Extract class
        * Extract subclass
        * Extract interface
        """


@Smell.subaspect
class MethodSmell:
    """
    Code smells related to a method or function definition.

    Method-level code smells indicate poorly defined method and or
    functions (too long method or functions, or functions with too many
    parameters) in your source code.
    """
    class docs:
        example = """
        def do_nothing(var1, var2, var3, var4, var5, var6, var7):
            pass
        """
        example_language = 'Python'
        importance_reason = """
        Make your functions and methods unambiguous, easy to read and debug
        by reducing the number of parameters and length of your methods and
        functions.
        """
        fix_suggestions = """
        A fix for this would simply consist of redefining the functions
        (and or method), making them shorter and reducing the number of
        parameters (maybe by creating more functions or using libraries).
        """


@MethodSmell.subaspect
class MethodLength:
    """
    Number of lines of code in a function or method definition.

    Depending on the value of `method_length_count_comment`,
    comments are considered. The `rule of 30
    <https://dzone.com/articles/rule-30-%E2%80%93-when-method-class-or>`_
    suggests that the maximum number of lines for a method is 30. ``PMD``
    defines a default value of 100 lines per method, `checkstlye`` 150 and
    60 (when comments are not considered), ``rubocop`` 10.
    """
    class docs:
        example = """
        def _is_positive(num):
            if num > 0:
                return True
            else:
                return False

        # This function can be defined as follow:

        def is_positive(num):
            return num > 0
        """
        example_language = 'Python'
        importance_reason = """
        It is really important is to stay DRY ("Don't Repeat Yourself") and
        respect separation of concerns. Long methods are sometimes faster and
        easier to write, and don't lead to maintenance problems. But most of
        the time they are an easy way to detect that something *may* be wrong
        in the code, and that special care is required while maintaining it.
        """
        fix_suggestions = """
        Refactoring methods into smaller more generic methods, making code more
        compact by using inline and language specific syntax tricks,
        implementing methods or functions to avoid redundant operation(s),
        making use of methods provided in libraries rather than reimplementing
        them.
        """
    max_method_length = Taste[int](
        'Represents the max number of lines for a method or a function\'s'
        'definition.',
        (10, 30, 50, 60, 100), default=30)
    method_length_count_comment = Taste[bool](
        'Allows when set to `True` to considered comments while calculating'
        'methods\' length.',
        (30, 60, 100, 150), default=60)


@MethodSmell.subaspect
class ParameterListLength:
    """
    Number of parameter a function or method has.

    In the book "Code Complete", ISBN 0735619670 it is suggested that the
    maximum number of parameter per function or method is 7; ``rubocop``
    suggests 5.
    """
    class docs:
        example = """
        def func(a, b, c, d, e, f, g, h, i, j, k):
            pass
        """
        example_language = 'Python'
        importance_reason = """
        Methods that take too many parameters are difficult to read, maintain
        and work with, and callers of those method often have an awkward time
        assembling all of the data and the resulting code is usually not pretty.
        """
        fix_suggestions = """
        This can be fixed by:

        Instead of passing a group of data received from an object as
        parameters, pass the object itself to the method.
        Sometimes you can merge several parameters into a single object etc.
        """
    max_parameters = Taste[int](
        'Represents the max number of parameters for a function or a method.',
        (5, 7), default=7)


@ClassSmell.subaspect
class DataClump:
    """
    Identical groups of variables found in many different part of a program.
    """
    class docs:
        example = """
        public static void main(String args[]) {

            String firstName = args[0];
            String lastName = args[1];
            Integer age = new Integer(args[2]);
            String gender = args[3];
            String occupation = args[4];
            String city = args[5];

            welcomeNew(firstName,lastName,age,gender,occupation,city);
        }

        public static void welcomeNew(String firstName, String lastName,
                              Integer age, String gender,
                              String occupation, String city){

            System.out.printf(
                "Welcome %s %s, a %d-year-old %s from %s who works as a %s",
                firstName, lastName, age, gender, city, occupation
            );
        }
        """
        example_language = 'Java'
        importance_reason = """
        Data clumps make code difficult to read, understand, and reuse.
        It also spoils their architecture.
        """
        fix_suggestions = """
        Formally group the different variables together into a single object.
        """


@ClassSmell.subaspect
class ClassSize:
    """
    Class size refers to the size of a class.

    A class's size is based on:

        * the number of fields it contains,
        * the number of methods,
        * and the number of lines of code.
    """
    class docs:
        example = """
        // This is large class given that the `max_class_length` is 20

        public class Employee
        {
            private float salary;
            private float bonusPercentage;
            private EmployeeType employeeType;
            public Employee(float salary, float bonusPercentage,
                            EmployeeType employeeType)
            {
                this.salary = salary;
                this.bonusPercentage = bonusPercentage;
                this.employeeType = employeeType;
            }
            public float CalculateSalary()
            {
                switch (employeeType)
                {
                    case EmployeeType.Worker:
                        return salary;
                    case EmployeeType.Supervisor:
                        return salary + (bonusPercentage * 0.5F);
                    case EmployeeType.Manager:
                        return salary + (bonusPercentage * 0.7F);
                }
                return 0.0F;
            }
        }
        """
        example_language = 'Java'
        importance_reason = """
        Refactoring large classes spares developers from the need to remember
        a large number of attributes and methods. Splitting these classes
        avoids duplication, and makes the code shorter and easier to maintain.
        Sometimes a Lazy Class (a class that does too little) is created in
        order to delineate intentions for future development, In this case,
        try to maintain a balance between clarity and simplicity in your code;
        and they should be deleted if they serve no purpose.
        """
        fix_suggestions = """
        Usually splitting up large classes into other classes or giving in
        line Class treatment to component that are near useless can solve this
        problem.
        """


@ClassSize.subaspect
class ClassLength:
    """
    Number of lines of code in class' definition.
    """
    class docs:
        example = """
        # Here is an example of a large class (in terms of number of lines) if
        # we assume that the maximum number of lines per class defintion is 10

        class Student:
            def __init__(self, first_name, last_name, dob,
                         matricule, school, faculty, department,
                         level, courses):
                self.first_name = first_name
                self.last_name = last_name
                self.dob = dob
                self.matricule = matricule
                self.school = school
                self.faculty = faculty
                self.department = department
                self.level = level
                self.courses = courses
        """
        example_language = 'Python 3'
        importance_reason = """
        Too large classes are difficult to read and maintain, and can easily
        introduce bugs or duplication in our code base.
        """
        fix_suggestions = """
        Usually splitting up those classes into other classes solves the
        problem.
        """
    max_class_length = Taste[int](
        'Represents the max number of lines for a class\'s definition.',
        (999, 900), default=900)


@ClassSize.subaspect
class ClassConstants:
    """
    Number of constants in a class.
    """
    class docs:
        example = """
        // Here is an example of class with too many constants if we assume
        // that the maximum number of constants per class is 4

        class Aclass {
            final public int a = 1;
            final public int b = 2;
            final public String c = "coala";
            final public String d = "aspectsYEAH";
            final public Boolean e = true;

            public Aclass(){}
        }
        """
        example_language = 'Java'
        importance_reason = """
        Avoids having too many constants to spare developers from the neeed
        to remember too many of them.
        """
        fix_suggestions = """
        `ClassConstants` issues can be fixed by using data classes.
        """
    max_constants = Taste[int](
        'Represents the max number of constants for a class',
        (3,), default=3)


@ClassSize.subaspect
class ClassInstanceVariables:
    """
    Number of instance variables in a class.
    """
    class docs:
        example = """
        # Here is an example of a class with a large number of instance
        # variables if we assume that the maximun nubmer of instance variables
        # per class is 5.

        class Student:
            def __init__(self, first_name, last_name, dob,
                         matricule, school, faculty, department,
                         level, courses):
                self.first_name = first_name
                self.last_name = last_name
                self.dob = dob
                self.matricule = matricule
                self.school = school
                self.faculty = faculty
                self.department = department
                self.level = level
                self.courses = courses
        """
        example_language = 'Python 3'
        importance_reason = """
        Refactoring these classes spares developers from the need to remember
        a large number of attributes.
        """
        fix_suggestions = """
        Usually splitting up those classes into other classes solves the
        problem.
        """
    max_instance_variables = Taste[int](
        'Represents the max number of instance variables for a class',
        (3,), default=3)


@ClassSize.subaspect
class ClassMethods:
    """
    Number of class methods a class has.
    """
    class docs:
        example = """
        # Here is an example of a class with too many methods
        # Assuming that the maximum per class is 5.

        class AClass:
            def x(self): pass
            def y(self): pass
            def z(self): pass
            def p(self): pass
            def q(self): pass
            def r(self): pass
        """
        example_language = 'Python 3'
        importance_reason = """
        Refactoring these classes spares developers from the need to remember
        a large number of methods.
        """
        fix_suggestions = """
        Usually splitting up those classes into other classes solves the
        problem.
        """
    max_methods = Taste[int](
        'Represents the max number of methods for a class',
        (3,), default=3)


@ClassSmell.subaspect
class FeatureEnvy:
    """
    Classes that excessively use methods of other classes.
    """
    class docs:
        example = """
        public class Phone {
            private final String unformattedNumber;
            public Phone(String unformattedNumber) {
                this.unformattedNumber = unformattedNumber;
            }
            public String getAreaCode() {
                return unformattedNumber.substring(0,3);
            }
            public String getPrefix() {
                return unformattedNumber.substring(3,6);
            }
            public String getNumber() {
                return unformattedNumber.substring(6,10);
            }
        }
        public class Customer…
            private Phone mobilePhone;
            public String getMobilePhoneNumber() {
                return "(" +
                    mobilePhone.getAreaCode() + ") " +
                    mobilePhone.getPrefix() + "-" +
                    mobilePhone.getNumber();
            }
        }
        """
        example_language = 'Java'
        importance_reason = """
        This smell may occur after fields are moved to a data class; which
        makes the code less readable, and difficult to debug.
        """
        fix_suggestions = """
        Move the operations on data to the newly defined class(given that
        the fields of one class were moved to this class) as well.
        """


@Smell.subaspect
class Naming:
    """
    `Naming` refers to the naming conventions to use for identifiers.
    """
    class docs:
        example = """
        dummyVar = None  # lowerCamelCase naming convention
        DummyVar = None  # UpperCamelCase naming convention
        dummy_var = None  # snake_case naming convention
        """
        example_language = 'Python 3'
        importance_reason = """
        Consistent use of naming convention, make the code easy to read
        and debug.
        """
        fix_suggestions = """
        Use the appropriate naming convention for each data type.
        """

    variable_naming_convention = Taste[str](
        'Naming convention to use for variables\'s identifiers',
        ('lowerCamelCase', 'snake_case', 'kebab-case', 'UpperCamelCase'),
        default='snake_case')
    function_naming_convention = Taste[str](
        'Naming convention to use for functions\'s or methods\'s identifiers',
        ('lowerCamelCase', 'snake_case', 'kebab-case', 'UpperCamelcase'),
        default='snake_case')
    class_naming_convention = Taste[str](
        'Naming convention to use for classes\'s identifiers',
        ('lowerCamelCase', 'snake_case', 'kebab-case', 'UpperCamelCase'),
        default='UpperCamelCase')
    max_identifier_length = Taste[int](
        'The maximum number of character for an identifier.',
        (31,), default=31)


@Smell.subaspect
class Complexity:
    """
    Complexity of a code based on different complexity metrics.
    """
    class docs:
        example = """
        * McCabe's complexity
        * Halstead complexity
        * Elshof complexity
        * Data complexity
        etc...

        Here is an example of complex code:
        https://github.com/sympy/sympy/blob/master/sympy/solvers/solvers.py
        """
        example_language = 'reStructuredText'
        importance_reason = """
        Complex programs are difficult to read and maintain. Reducing a code's
        complexity improves its organization.
        """
        fix_suggestions = """
        Implementing simple methods, avoiding too many branches, avoiding too
        much multilevel inheritance etc... can fix this.
        """


@Complexity.subaspect
class CylomaticComplexity:
    """
    Number of linearly independent paths through a program's source code.

    The `Cyclomatic complexity
    <https://wikipedia.org/wiki/Cyclomatic_complexity>`_  was developed by
    Thomas J. McCabe in 1976 and it is based on a control flow representation
    of the program.
    """
    class docs:
        example = """
        // The cyclomatic complexity of this program is 4.

        int foo (int a, int b) {
            if (a > 17 && b < 42 && a+b < 55) {
                return 1;
            }
            return 2;
        }
        """
        example_language = 'C'
        importance_reason = """
        Very complex code are difficult to read, debug and maintain.
        It is always a good idea to keep things as simple as possible.
        """
        fix_suggestions = """
        This can be solved by breaking down complex functions into smaller
        onces.
        """
    cyclomatic_complexity = Taste[int](
        'This the maximum number of embedded branches or embedded loops'
        ' allowed.',
        (6,), default=6)


@Complexity.subaspect
class MaintainabilityIndex:
    """
    Software metric which measure how maintainable is a program.

    The `maintainablity index
    <www.projectcodemeter.com/cost_estimation/help/GL_maintainability.htm>`_
    is always in the range 0-100 and is ranked
    as follow:

    * A `MI` in the range 0-9 maps to a code extremely difficult to maintain.
    * A `MI` in the range 10-19 maps to a maintainable code.
    * A `MI` in the range 20-100 maps to a code highly maintainable.
    """
    class docs:
        example = """
        '''
        The maintainability index for the following piece of code is 100.
        '''
        def preorder(node):
            if tree:
                print(node.key)
                preorder(node.left)
                preorder(node.right)
        """
        example_language = 'Python'
        importance_reason = """
        Complex codes are difficult to maintain.
        """
        fix_suggestions = """
        This can be solved by writing simpler functions and methods.
        """
    maintainability_index = Taste[int](
        'Maintainability index of your code',
        tuple(range(100)), default=10)



class AspectTypeError(TypeError):
    """
    This error is raised when an object is not an ``aspectclass`` or an
    instance of ``aspectclass``
    """

    def __init__(self, item):
        self.item = item
        message = ('{} is not an aspectclass or an instance of an '
                   'aspectclass'.format(repr(self.item)))
        super().__init__(message)


class AspectLookupError(LookupError):
    """
    Error raised when trying to search aspect.
    """

    def __init__(self, aspectname, message=None):
        self.aspectname = aspectname
        if message is None:
            message = ('Error when trying to search aspect named {}'
                       .format(repr(aspectname)))
        super().__init__(message)


class AspectNotFoundError(AspectLookupError):
    """
    No aspect found.
    """

    def __init__(self, aspectname):
        message = ('No aspect named {}'.format(repr(aspectname)))
        super().__init__(aspectname, message)


class MultipleAspectFoundError(AspectLookupError):
    """
    Multiple aspect are found.
    """

    def __init__(self, aspectname, other_aspects):
        self.other_aspects = other_aspects
        message = ('Multiple aspects named {}. Choose from {}'.format(
            repr(aspectname),
            repr(sorted(other_aspects, key=lambda a: a.__qualname__))))
        super().__init__(aspectname, message)



import logging
import traceback

from coalib.misc import Constants
from coalib.output.printers.LOG_LEVEL import LOG_LEVEL

from pkg_resources import VersionConflict


def get_exitcode(exception, log_printer=None):
    if isinstance(exception, KeyboardInterrupt):  # Ctrl+C
        print('Program terminated by user.')
        exitcode = 130
    elif isinstance(exception, EOFError):  # Ctrl+D
        print('Found EOF. Exiting gracefully.')
        exitcode = 0
    elif isinstance(exception, SystemExit):
        exitcode = exception.code
    elif isinstance(exception, VersionConflict):
        log_message = Constants.VERSION_CONFLICT_MESSAGE % str(exception.req)
        log_exception(log_message, exception)
        exitcode = 13
    elif isinstance(exception, BaseException):
        log_exception(Constants.CRASH_MESSAGE, exception)
        exitcode = 255
    else:
        exitcode = 0

    return exitcode


def log_exception(message, exception, log_level=LOG_LEVEL.ERROR):
    logging.log(log_level, message)
    traceback_str = '\n'.join(
        traceback.format_exception(type(exception),
                                   exception,
                                   exception.__traceback__))
    logging.info('Exception was:\n' + traceback_str)



import logging
import time
import os

from coala_utils.decorators import enforce_signature
from coalib.misc.CachingUtilities import (
    pickle_load, pickle_dump, delete_files)


class FileCache:
    """
    This object is a file cache that helps in collecting only the changed
    and new files since the last run. Example/Tutorial:

    >>> import logging
    >>> import copy, time
    >>> logging.getLogger().setLevel(logging.CRITICAL)

    To initialize the cache create an instance for the project:

    >>> cache = FileCache(None, "test", flush_cache=True)

    Now we can track new files by running:

    >>> cache.track_files(["a.c", "b.c"])

    Since all cache operations are lazy (for performance), we need to
    explicitly write the cache to disk for persistence in future uses:
    (Note: The cache will automatically figure out the write location)

    >>> cache.write()

    Let's go into the future:

    >>> time.sleep(1)

    Let's create a new instance to simulate a separate run:

    >>> cache = FileCache(None, "test", flush_cache=False)

    >>> old_data = copy.deepcopy(cache.data)

    We can mark a file as changed by doing:

    >>> cache.untrack_files({"a.c"})

    Again write to disk after calculating the new cache times for each file:

    >>> cache.write()
    >>> new_data = cache.data

    Since we marked 'a.c' as a changed file:

    >>> "a.c" not in cache.data
    True
    >>> "a.c" in old_data
    True

    Since 'b.c' was untouched after the second run, its time was updated
    to the latest value:

    >>> old_data["b.c"] < new_data["b.c"]
    True
    """

    @enforce_signature
    def __init__(
            self,
            log_printer,
            project_dir: str,
            flush_cache: bool=False):
        """
        Initialize FileCache.

        :param log_printer: An object to use for logging.
        :param project_dir: The root directory of the project to be used
                            as a key identifier.
        :param flush_cache: Flush the cache and rebuild it.
        """
        self.project_dir = project_dir
        self.current_time = int(time.time())

        cache_data = pickle_load(None, project_dir, {})
        last_time = -1
        if 'time' in cache_data:
            last_time = cache_data['time']
        if not flush_cache and last_time > self.current_time:
            logging.warning('It seems like you went back in time - your system '
                            'time is behind the last recorded run time on this '
                            'project. The cache will be force flushed.')
            flush_cache = True

        self.data = cache_data.get('files', {})
        if flush_cache:
            self.flush_cache()

        # store the files to be untracked and then untrack them in the end
        # so that an untracked file is not tracked again by mistake in a
        # later section (which will happen if that file doesn't yield a
        # result in that section).
        self.to_untrack = set()

    def flush_cache(self):
        """
        Flushes the cache and deletes the relevant file.
        """
        self.data = {}
        delete_files(None, [self.project_dir])
        logging.debug('The file cache was successfully flushed.')

    def __enter__(self):
        return self

    def write(self):
        """
        Update the last run time on the project for each file
        to the current time. Using this object as a contextmanager is
        preferred (that will automatically call this method on exit).
        """
        for file in self.to_untrack:
            if file in self.data:
                del self.data[file]
        for file_name in self.data:
            self.data[file_name] = self.current_time
        pickle_dump(
            None,
            self.project_dir,
            {'time': self.current_time, 'files': self.data})

    def __exit__(self, type, value, traceback):
        """
        Update the last run time on the project for each file
        to the current time.
        """
        self.write()

    def untrack_files(self, files):
        """
        Removes the given files from the cache so that they are no longer
        considered cached for this and the next run.

        :param files: A set of files to remove from cache.
        """
        self.to_untrack.update(files)

    def track_files(self, files):
        """
        Start tracking files given in ``files`` by adding them to the
        database.

        :param files: A set of files that need to be tracked.
                      These files are initialized with their last
                      modified tag as -1.
        """
        for file in files:
            if file not in self.data:
                self.data[file] = -1

    def get_uncached_files(self, files):
        """
        Returns the set of files that are not in the cache yet or have been
        untracked.

        :param files: The list of collected files.
        :return:      A set of files that are uncached.
        """
        if self.data == {}:
            # The first run on this project. So all files are new
            # and must be returned irrespective of whether caching is turned on.
            return files
        else:
            return {file
                    for file in files
                    if (file not in self.data or
                        int(os.path.getmtime(file)) > self.data[file])}



from collections import namedtuple

from coala_utils.decorators import generate_eq, generate_repr
from coalib.results.TextRange import TextRange
from functools import lru_cache


@generate_repr()
@generate_eq('documentation', 'language', 'docstyle',
             'indent', 'marker', 'position')
class DocumentationComment:
    """
    The DocumentationComment holds information about a documentation comment
    inside source-code, like position etc.
    """
    Parameter = namedtuple('Parameter', 'name, desc')
    ExceptionValue = namedtuple('ExceptionValue', 'name, desc')
    ReturnValue = namedtuple('ReturnValue', 'desc')
    Description = namedtuple('Description', 'desc')
    top_padding = 0
    bottom_padding = 0
    docstring_type = 'others'

    def __init__(self, documentation, docstyle_definition,
                 indent, marker, position):
        """
        Instantiates a new DocumentationComment.

        :param documentation:
            The documentation text.
        :param docstyle_definition:
            The ``DocstyleDefinition`` instance that defines what docstyle is
            being used in the documentation.
        :param indent:
            The string of indentation used in front of the first marker of the
            documentation.
        :param marker:
            The three-element tuple with marker strings, that identified this
            documentation comment.
        :param position:
            The starting ``TextPosition`` of the documentation.
        """
        self.documentation = documentation
        self.docstyle_definition = docstyle_definition
        self.indent = '' if indent is None else indent
        self.marker = ('', '', '') if marker is None else marker
        self.position = position
        self.range = None if position is None else TextRange.from_values(
            position.line,
            position.column,
            position.line + self.assemble().count('\n'),
            len(self.assemble()) - self.assemble().rfind('\n'))

    def __str__(self):
        return self.documentation

    @property
    def language(self):
        return self.docstyle_definition.language

    @property
    def docstyle(self):
        return self.docstyle_definition.docstyle

    @property
    def metadata(self):
        return self.docstyle_definition.metadata

    def parse(self):
        """
        Parses documentation independent of language and docstyle.

        :return:
            The list of all the parsed sections of the documentation. Every
            section is a namedtuple of either ``Description`` or ``Parameter``
            or ``ReturnValue``.
        :raises NotImplementedError:
            When no parsing method is present for the given language and
            docstyle.
        """
        if self.language == 'python' and self.docstyle == 'default':
            return self._parse_documentation_with_symbols(
                (':param ', ':'), (':raises ', ':'), ':return:')
        elif self.language == 'python' and self.docstyle == 'doxygen':
            return self._parse_documentation_with_symbols(
                ('@param ', ' '), ('@raises ', ' '), '@return ')
        elif self.language == 'java' and self.docstyle == 'default':
            return self._parse_documentation_with_symbols(
                ('@param  ', ' '), ('@raises  ', ' '), '@return ')
        elif self.language == 'golang' and self.docstyle == 'golang':
            # golang does not have param, return markers
            return self.documentation.splitlines(keepends=True)
        else:
            raise NotImplementedError(
                'Documentation parsing for {0.language!r} in {0.docstyle!r}'
                ' has not been implemented yet'.format(self))

    def _parse_documentation_with_symbols(self,
                                          param_identifiers,
                                          exception_identifiers,
                                          return_identifiers):
        """
        Parses documentation based on parameter, exception and return symbols.

        :param param_identifiers:
            A tuple of two strings with which a parameter starts and ends.
        :param exception_identifiers:
            A tuple of two strings with which an exception starts and ends.
        :param return_identifiers:
            The string with which a return description starts.
        :return:
            The list of all the parsed sections of the documentation. Every
            section is a named tuple of either ``Description``, ``Parameter``,
            ``ExceptionValue`` or ``ReturnValue``.
        """
        lines = self.documentation.splitlines(keepends=True)

        parse_mode = self.Description

        cur_param = ''

        desc = ''
        parsed = []

        for line in lines:

            stripped_line = line.strip()

            if stripped_line.startswith(param_identifiers[0]):
                parse_mode = self.Parameter
                # param_offset contains the starting column of param's name.
                param_offset = line.find(
                    param_identifiers[0]) + len(param_identifiers[0])
                # splitted contains the whole line from the param's name,
                # which in turn is further divided into its name and desc.
                splitted = line[param_offset:].split(param_identifiers[1], 1)
                # parser breaks if param_identifiers[1] is not present.
                # This checks for space and then splits the line accordingly
                # to extract param's name and desc.
                if len(splitted) == 1:
                    splitted = line[param_offset:].split(' ', 1)
                cur_param = splitted[0].strip()

                param_desc = splitted[1]
                # parsed section is added to the final list.
                parsed.append(self.Parameter(name=cur_param, desc=param_desc))

            elif stripped_line.startswith(exception_identifiers[0]):
                parse_mode = self.ExceptionValue
                exception_offset = line.find(
                    exception_identifiers[0]) + len(exception_identifiers[0])
                splitted = line[exception_offset:].split(
                    exception_identifiers[1], 1)
                # parser breaks if exception_identifiers[1] is not present.
                # This checks for space and then splits the line accordingly
                # to extract exception's name and desc.
                if len(splitted) == 1:
                    splitted = line[exception_offset:].split(' ', 1)
                cur_exception = splitted[0].strip()

                exception_desc = splitted[1]
                parsed.append(self.ExceptionValue(
                    name=cur_exception, desc=exception_desc))

            elif stripped_line.startswith(return_identifiers):
                parse_mode = self.ReturnValue
                return_offset = line.find(
                    return_identifiers) + len(return_identifiers)
                retval_desc = line[return_offset:]
                parsed.append(self.ReturnValue(desc=retval_desc))

            # These conditions will take care if the parsed section
            # descriptions are not on the same line as that of it's
            # name. Further, adding the parsed section to the final list.
            elif parse_mode == self.ReturnValue:
                retval_desc += line
                parsed.pop()
                parsed.append(self.ReturnValue(desc=retval_desc))

            elif parse_mode == self.ExceptionValue:
                exception_desc += line
                parsed.pop()
                parsed.append(self.ExceptionValue(
                    name=cur_exception, desc=exception_desc))

            elif parse_mode == self.Parameter:
                param_desc += line
                parsed.pop()
                parsed.append(self.Parameter(name=cur_param, desc=param_desc))

            else:
                desc += line
                # This is inside a try-except for cases where the list
                # is empty and has nothing to pop.
                try:
                    parsed.pop()
                except IndexError:
                    pass
                parsed.append(self.Description(desc=desc))

        return parsed

    @classmethod
    def from_metadata(cls, doccomment, docstyle_definition,
                      marker, indent, position):
        r"""
        Assembles a list of parsed documentation comment metadata.

        This function just assembles the documentation comment
        itself, without the markers and indentation.

        >>> from coalib.bearlib.languages.documentation.DocumentationComment \
        ...     import DocumentationComment
        >>> from coalib.bearlib.languages.documentation.DocstyleDefinition \
        ...     import DocstyleDefinition
        >>> from coalib.results.TextPosition import TextPosition
        >>> Description = DocumentationComment.Description
        >>> Parameter = DocumentationComment.Parameter
        >>> python_default = DocstyleDefinition.load("python3", "default")
        >>> parsed_doc = [Description(desc='\nDescription\n'),
        ...               Parameter(name='age', desc=' Age\n')]
        >>> str(DocumentationComment.from_metadata(
        ...         parsed_doc, python_default,
        ...         python_default.markers[0], '    ',
        ...         TextPosition(0, 0)))
        '\nDescription\n:param age: Age\n'

        :param doccomment:
            The list of parsed documentation comment metadata.
        :param docstyle_definition:
            The ``DocstyleDefinition`` instance that defines what docstyle is
            being used in a documentation comment.
        :param marker:
            The markers to be used in the documentation comment.
        :param indent:
            The indentation to be used in the documentation comment.
        :param position:
            The starting position of the documentation comment.
        :return:
            A ``DocumentationComment`` instance of the assembled documentation.
        """
        assembled_doc = ''
        for section in doccomment:
            section_desc = section.desc.splitlines(keepends=True)

            if isinstance(section, cls.Parameter):
                assembled_doc += (docstyle_definition.metadata.param_start +
                                  section.name +
                                  docstyle_definition.metadata.param_end)

            elif isinstance(section, cls.ExceptionValue):
                assembled_doc += (docstyle_definition.metadata.exception_start
                                  + section.name
                                  + docstyle_definition.metadata.exception_end)

            elif isinstance(section, cls.ReturnValue):
                assembled_doc += docstyle_definition.metadata.return_sep

            assembled_doc += ''.join(section_desc)

        return DocumentationComment(assembled_doc, docstyle_definition, indent,
                                    marker, position)

    # we need to cache this function so as to construct full `self.range`
    @lru_cache(maxsize=1)
    def assemble(self):
        """
        Assembles parsed documentation to the original documentation.

        This function assembles the whole documentation comment, with the
        given markers and indentation.
        """
        lines = self.documentation.splitlines(keepends=True)
        assembled = self.indent + self.marker[0]
        if len(lines) == 0:
            return self.marker[0] + self.marker[2]
        assembled += lines[0]
        assembled += ''.join('\n' if line == '\n' and not self.marker[1]
                             else self.indent + self.marker[1] + line
                             for line in lines[1:])
        assembled = (assembled if self.marker[1] == self.marker[2] else
                     (assembled +
                      (self.indent if lines[-1][-1] == '\n' else '') +
                      self.marker[2]))
        assembled = ('\n' * self.top_padding + assembled +
                     '\n' * self.bottom_padding)
        return assembled


class MalformedComment:
    """
    The MalformedComment holds information about the errors generated by the
    DocumentationExtraction, DocumentationComment, DocstyleDefinition and
    DocBaseClass.

    When these classes are unable to parse certain docstrings, an instance
    of MalformedComment will be returned instead of DocumentationComment.
    """

    def __init__(self, message, line):
        """
        Instantiate a MalformedComment, which contains the information about
        the error: a message explaining the behaviour and a line no where the
        error has occured.

        :param message:
            Contains the message about the error.
        :param line:
            Contains the current line of the docstring where the error has
            occured.
        """
        self.message = message
        self.line = line



from pygments.lexers import guess_lexer_for_filename
from pyprint.ConsolePrinter import ConsolePrinter

from coalib.results.result_actions.ResultAction import ResultAction
from coalib.results.result_actions.ShowPatchAction import ShowPatchAction
from coalib.output.printers.LogPrinter import LogPrinter
from coalib.parsing.FilterHelper import apply_filters
from coalib.coala_modes import mode_normal
from coalib.parsing.DefaultArgParser import default_arg_parser


OBJECT_INDEX = FILENAME_INDEX = 0
DEFAULT_BEAR = 'SpaceConsistencyBear'


def show_possibilities(console_printer, i, action):
    console_printer.print('[{:>4}]  {}. Apply patch (\'{}\')'.format('', i,
                                                                     action))


def create_arg_parser(files, bears):
    """
    A function that generates a `default_arg_parser`.

    :param files: A list that contains filenames.
    :param bears: A list that contains name of bears.
    :return:      An object of type `default_arg_parser`.
    """
    args = default_arg_parser().parse_args()
    args.files = files
    args.bears = bears
    args.default_actions = '*: ApplyPatchAction'

    return args


def filter_bears(language):
    """
    Filter bears by language.

    :param language: The language to filter with.
    :return:         A list of bears.
    """
    return list(apply_filters([['language', language]], None)[0]['cli'])


def find_language(filename):
    """
    Find the language used in `filename`.

    :param filename: The name of the file.
    :return:         The language used.
    """

    return guess_lexer_for_filename(filename, 'Error, no file '
                                    'found').name


class DefaultBear():
    def __init__(self):
        self.name = DEFAULT_BEAR


class GeneratePatchesAction(ResultAction):
    SUCCESS_MESSAGE = 'Patch generated successfully.'

    is_applicable = staticmethod(ShowPatchAction.is_applicable)

    def apply(self,
              result,
              original_file_dict,
              file_diff_dict):
        """
        (G)enerate patches
        """

        console_printer = ConsolePrinter()
        log_printer = LogPrinter()
        to_filename = sorted(result.diffs.items())[OBJECT_INDEX][FILENAME_INDEX]

        filtered_bears = filter_bears(find_language(to_filename))
        filtered_bears.insert(0, DefaultBear())
        possible_options = [b.name for b in filtered_bears]

        console_printer.print('[{:>4}] *0. Do Nothing'.format(''))

        # Let the user choose a bear that wants to apply on the files
        for i, action in enumerate(possible_options, 1):
            show_possibilities(console_printer, i, action)

        choose_action = str(input('[{:>4}]  Enter a number: '.format('')))
        if choose_action is '' or choose_action is '0':
            return False

        choose_action = int(choose_action)
        chosen_bear = [possible_options[choose_action - 1]]

        return mode_normal(console_printer, log_printer, create_arg_parser([
            to_filename], chosen_bear), debug=False)
